#define BYTE unsigned char
#define WORD unsigned short
#include "..\test\system.inc"
#include "pci.h"
#include "..\test\ports.inc"

#define DEBUG //
#define  calculate_adress(bus,device)  ((bus  << 16) + (device << 8) + 0x80000000)

int   testbit(BYTE bits, BYTE num);
void  setbit(BYTE *bits, BYTE num);
int   pci_test_device(DWORD adress);
/*DWORD scan_first(BYTE *bus_bit_map);*/
/*void  scan_second(BYTE *pbm);*/
void  scan_bus(DWORD bus_nr);
void  scan_device(DWORD adress);
void  init_driver();

BYTE  pbm[32] = {0}; /*PCI BUS MAP*/

DWORD           *global_pci_dev_info_size = 0;
PCI_DEVICE_INFO *global_pci_dev_info      = 0;
DWORD            global_dev_pos = 0;

DWORD global_user_count = 0;
DWORD shutdownmode;
DWORD *shutdowntask;

ports_tportintv pcidrv_port_intervall;

/******************************************************************************
  * pci_test_device testet ob sich ein Gert an der
  * angegebenen Adresse befindet und gibt als Rckgabewert
  * entweder 0 fr kein Gert oder 1 fr Gert gefunden
  *****************************************************************************/

int pci_test_device(DWORD adress)
{
  DWORD result = 0x00000000;
  outportd(0xcf8,adress);
  result = inportd(0xcfc);
  return ((result & 0xffff) != 0xffff);
}

/******************************************************************************
  * setbit setzt in einem BYTE (8 Bit) ein bit 
  * an der gewnschten Stelle
  *****************************************************************************/

void setbit(BYTE *bits, BYTE num)
{
  if(num > 7) return;
  (*bits) |= ( 1 << num);
}

/******************************************************************************
  * testbit testet ob an der angegebenen Stelle
  * im BYTE ein Bit gesetzt ist oder nicht
  *****************************************************************************/

int testbit(BYTE bits, BYTE num)
{
  return ((bits & ( 1 << num)) > 0);
}

/******************************************************************************
  * scan_first sucht alle gerte und markiert den 
  * entsprechenden Bus in einer 32 * 8 - BitMap.
  * Desweiteren zhlt die funktion auch gleich wieviel 
  * Gerte sich in dem Computer befinden. 
  * Das wird gemacht um spter 1. nicht mehr alle Buse scannen 
  * zu mssen und 2. damit man weis fr wieviel Speicher 
  * allokiert werden muss.
  *****************************************************************************/
DWORD scan_first()
{
 DWORD cnt = 0;
 WORD bus_nr = 0, device = 0;
 for(bus_nr = 0; bus_nr < 256; ++bus_nr)
     for(device = 0; device < 256; ++device)
     {
        if(pci_test_device(calculate_adress(bus_nr,device)))
        {
           setbit(&pbm[bus_nr/8],bus_nr % 8);
           cnt++;
        }
     }
  return cnt;
}

/******************************************************************************
  * scan_second checkt welche Busse 
  * in der BitMap pbm markiert sind
  * und ruft fr jeden dieser Buse scan_bus() auf
  *****************************************************************************/
void scan_second()
{
  DWORD i = 0, j = 0;
  for(i = 0; i < 32; ++i)
  {
    for(j = 0; j < 8; ++j)
       if(testbit(pbm[i],j))
        scan_bus(j * i);
  }
}

/******************************************************************************
  * scan_device liest aus einem Gert alle 
  * fr PCI_DEVICE_INFO bentigten Informationen.
  * und speichert diese in das global_pci_dev_info
  * 
  *****************************************************************************/
void scan_device(DWORD adress)
{

  if(pci_test_device(adress) && global_dev_pos < *global_pci_dev_info_size)
  {
    DEBUG("checking PCI Device at: 0x%x\n",adress);
    DWORD result = 0xFFFFFFFF;

    global_pci_dev_info[global_dev_pos].adress = adress;
    global_pci_dev_info[global_dev_pos].flags  = 0x00000000;

    outportd(0xcf8,adress);
    result = inportd(0xcfc);
    global_pci_dev_info[global_dev_pos].ids    = result;
    result = inportd(0xcfc + 0x08);
    global_pci_dev_info[global_dev_pos].class_code_rev = result;
    global_dev_pos++;
  }
}
/******************************************************************************
 ** scan_bus sucht Gert auf dem angegebenen Bus
 ******************************************************************************/
void scan_bus(DWORD bus_nr)
{
  DWORD i = 0;
  for(i = 0; i < 256; ++i)
  {
    scan_device(0x80000000 + ( bus_nr << 16 ) + (i << 8));
  }

}

/******************************************************************************
  * init_driver initialisiert den Treiber
  * und reserviert den Speicher fr die Gerte Daten
  *****************************************************************************/

void init_driver()
{
  pcidrv_port_intervall.start = 0x3f8;
  pcidrv_port_intervall.ende  = 0xcff;
  pcidrv_port_intervall.next  = 0;
  pcidrv_port_intervall.last  = 0;
  addportintv(&pcidrv_port_intervall);

  DWORD device_count = 0;
  DWORD dev = 0;
  for(dev = 0; dev < 32; ++dev)
  {
    pbm[dev] = 0x00;
  }

  /** 1. Scannen der Gerte...
    */
  device_count = scan_first();

  /** Speicher auf dem Heap allokieren
    */
  global_pci_dev_info      = (PCI_DEVICE_INFO*)allocmem(sizeof(PCI_DEVICE_INFO)*device_count+1);
  global_pci_dev_info_size = (DWORD*)allocmem(sizeof(DWORD));

  /** bergabe der Anzahl der Gerte an die globale Heap Variable
    */
  *global_pci_dev_info_size = device_count;

  /** 2. Scannen der Gerte mit auslesen der informationen
    * (Holt sich die Variablen direkt da es globale sind.)
    */
  scan_second();
}
/******************************************************************************
  *
  ****************************************************************************/
void PASCAL deinit_pci_manager_driver()
{
  shutdowntask = system_gettaskhandle();
  shutdownmode = 1;
  if(global_user_count) system_disabletask(shutdowntask);
  system_signalshutdownok();
   /** Freigeben des benutzten speichers
   */
 if(global_pci_dev_info != 0)
       releasemem(global_pci_dev_info,(*global_pci_dev_info_size) * sizeof(PCI_DEVICE_INFO));
 if(global_pci_dev_info_size != 0)
       releasemem(global_pci_dev_info_size,sizeof(DWORD));

  removeportintv(&pcidrv_port_intervall);

  ports_done();
  system_done();

  system_destroyprocess(system_getprocesshandle(),0);
}
/******************************************************************************
  * Freigabe des benutzten Speichers und Kontrolle der Benutzung
  ****************************************************************************/
void PASCAL free_pci_manager_prot(DWORD x)
{
  switch(x)
  {
    case 1:
           global_user_count--;
           if((shutdownmode) && (!global_user_count))system_enabletask(shutdowntask);
           return;
    default:
           return;
  }
}

/******************************************************************************
  * Behandlung der Protokolle
  *****************************************************************************/

void *pci_manager_prot0[] = {
                             (void*)&deinit_pci_manager_driver,
                             (void*)&free_pci_manager_prot
                            };
void *pci_manager_prot1[] = {};

void * PASCAL dock_pci_manager(DWORD prot)
{
  switch(prot)
  {
    case 0:
            return pci_manager_prot0;

    case 1:
            global_user_count++;
            return pci_manager_prot1;

    default: return 0;
  }
}

/******************************************************************************
  * main startet die initialisierung des Treibers
  *****************************************************************************/
system_tdriver pci_managing_driver=
   {0,0,0x17,0,&dock_pci_manager,"PCI Management Treiber (0,1,2)"};

int main()
{
 system_init();
 ports_init();
 init_driver();

 system_adddriverc(&pci_managing_driver);
 system_destroyprocess(system_getprocesshandle(),0);

 return 0;
}
